/* vi: set sw=4 ts=4: */
/*
 * Licensed under GPLv2 or later, see file LICENSE in this tarball for details.
 */

#include "libbb.h"
#include "unarchive.h"


/**
 * Copy of bb_full_fd_action but with fadvise(DONTNEED) to stop the kernel from
 * filling the page cache with pages that are written out, and unlikely to be read
 * back in before the cache has wrapped anyway. We are keen to avoid pollution
 * of the page cache as it would force out binary pages, these are then required
 * to be read back in from loop0 (which is necessarily exceptionally slow).
 *
 * To achieve that, we advise the kernel:
 *  1. for every 4 Mbytes written, any blocks already written out can be freed
 *  2. on every file close, and every 32 Mbytes written, forcibly sync all data
 *     then free all blocks.
 *
 * We have to do the fdatasync because otherwise (current) linux will ignore the
 * fadvise on blocks that have not yet been written out, and will not remember the
 * call for future.
 * 
 * This whole method of doing is a work around only. The choice of 32 Mbyte was
 * done based on the siena release (128MB/Nav3). This will need altering for other
 * devices. There must be a better way of limiting linux's page cache use for
 * tar/openssl cases.
 */

#define SYNC_BLOCKS (32 * 1024 / (CONFIG_FEATURE_COPYBUF_KB))
static const int syncBlocks = SYNC_BLOCKS; // 32 Mbytes
static const int syncBlocksFadvise = SYNC_BLOCKS / 8;

static off_t bb_full_fd_action_dea(int src_fd, int dst_fd, off_t size)
{
    int syncCount = 0;

	int status = -1;
	off_t total = 0;
	char buffer[CONFIG_FEATURE_COPYBUF_KB * 1024];
	enum { buffer_size = sizeof(buffer) };

	if (src_fd < 0)
		goto out;

	if (!size) {
		size = buffer_size;
		status = 1; /* copy until eof */
	}

	while (1) {
		ssize_t rd;

		rd = safe_read(src_fd, buffer, size > buffer_size ? buffer_size : size);

		if (!rd) { /* eof - all done */
			status = 0;
			break;
		}
		if (rd < 0) {
			bb_perror_msg(bb_msg_read_error);
			break;
		}
		/* dst_fd == -1 is a fake, else... */
		if (dst_fd >= 0) {
			ssize_t wr = full_write(dst_fd, buffer, rd);

			syncCount = (syncCount + 1) % (8192);
			if (!(syncCount % syncBlocksFadvise)) {
			    // sync our write out to disk and throw it out of the cache, otherwise
			    // the fadvise will not do anything
			    if (!syncCount) {
				fdatasync(dst_fd);
			    }
			    posix_fadvise(dst_fd, 0,0,POSIX_FADV_DONTNEED);
			}

			if (wr < rd) {
				bb_perror_msg(bb_msg_write_error);
				break;
			}
		}
		total += rd;
		if (status < 0) { /* if we aren't copying till EOF... */
			size -= rd;
			if (!size) {
				/* 'size' bytes copied - all done */
				status = 0;
				break;
			}
		}
	}
 out:
	// note - this will penalise packages with lots of small files, but does guarantee
	// to limit our usage of the page cache for the good of large (> PND cache size)
	// downloads
	fdatasync(dst_fd);
	posix_fadvise(dst_fd, 0,0,POSIX_FADV_DONTNEED);
	return status ? -1 : total;
}


off_t FAST_FUNC bb_copyfd_size_dea(int fd1, int fd2, off_t size);
off_t FAST_FUNC bb_copyfd_size_dea(int fd1, int fd2, off_t size)
{
	if (size) {
		return bb_full_fd_action_dea(fd1, fd2, size);
	}
	return 0;
}

void FAST_FUNC bb_copyfd_exact_size_dea(int fd1, int fd2, off_t size);
void FAST_FUNC bb_copyfd_exact_size_dea(int fd1, int fd2, off_t size)
{
	off_t sz = bb_copyfd_size_dea(fd1, fd2, size);
	if (sz == size)
		return;
	if (sz != -1)
		bb_error_msg_and_die("short read");
	/* if sz == -1, bb_copyfd_XX already complained */
	xfunc_die();
}



void FAST_FUNC data_extract_all(archive_handle_t *archive_handle)
{
	file_header_t *file_header = archive_handle->file_header;
	int dst_fd;
	int res;

	if (archive_handle->ah_flags & ARCHIVE_CREATE_LEADING_DIRS) {
		char *name = xstrdup(file_header->name);
		bb_make_directory(dirname(name), -1, FILEUTILS_RECUR);
		free(name);
	}

	/* Check if the file already exists */
	if (archive_handle->ah_flags & ARCHIVE_EXTRACT_UNCONDITIONAL) {
		/* Remove the entry if it exists */
		if (((file_header->mode & S_IFMT) != S_IFDIR)
		 && (unlink(file_header->name) == -1)
		 && (errno != ENOENT)
		) {
			bb_perror_msg_and_die("cannot remove old file %s",
					file_header->name);
		}
	}
	else if (archive_handle->ah_flags & ARCHIVE_EXTRACT_NEWER) {
		/* Remove the existing entry if its older than the extracted entry */
		struct stat statbuf;
		if (lstat(file_header->name, &statbuf) == -1) {
			if (errno != ENOENT) {
				bb_perror_msg_and_die("cannot stat old file");
			}
		}
		else if (statbuf.st_mtime <= file_header->mtime) {
			if (!(archive_handle->ah_flags & ARCHIVE_EXTRACT_QUIET)) {
				bb_error_msg("%s not created: newer or "
					"same age file exists", file_header->name);
			}
			data_skip(archive_handle);
			return;
		}
		else if ((unlink(file_header->name) == -1) && (errno != EISDIR)) {
			bb_perror_msg_and_die("cannot remove old file %s",
					file_header->name);
		}
	}

	/* Handle hard links separately
	 * We identified hard links as regular files of size 0 with a symlink */
	if (S_ISREG(file_header->mode) && (file_header->link_target)
	 && (file_header->size == 0)
	) {
		/* hard link */
		res = link(file_header->link_target, file_header->name);
		if ((res == -1) && !(archive_handle->ah_flags & ARCHIVE_EXTRACT_QUIET)) {
			bb_perror_msg("cannot create %slink "
					"from %s to %s", "hard",
					file_header->name,
					file_header->link_target);
		}
	} else {
		/* Create the filesystem entry */
		switch (file_header->mode & S_IFMT) {
		case S_IFREG: {
			/* Regular file */
			dst_fd = xopen3(file_header->name, O_WRONLY | O_CREAT | O_EXCL,
							file_header->mode);
			bb_copyfd_exact_size_dea(archive_handle->src_fd, dst_fd, file_header->size);
			close(dst_fd);
			break;
		}
		case S_IFDIR:
			res = mkdir(file_header->name, file_header->mode);
			if ((res == -1) && (errno != EISDIR)
			 && !(archive_handle->ah_flags & ARCHIVE_EXTRACT_QUIET)
			) {
				bb_perror_msg("cannot make dir %s", file_header->name);
			}
			break;
		case S_IFLNK:
			/* Symlink */
			res = symlink(file_header->link_target, file_header->name);
			if ((res == -1)
			 && !(archive_handle->ah_flags & ARCHIVE_EXTRACT_QUIET)
			) {
				bb_perror_msg("cannot create %slink "
					"from %s to %s", "sym",
					file_header->name,
					file_header->link_target);
			}
			break;
		case S_IFSOCK:
		case S_IFBLK:
		case S_IFCHR:
		case S_IFIFO:
			res = mknod(file_header->name, file_header->mode, file_header->device);
			if ((res == -1)
			 && !(archive_handle->ah_flags & ARCHIVE_EXTRACT_QUIET)
			) {
				bb_perror_msg("cannot create node %s", file_header->name);
			}
			break;
		default:
			bb_error_msg_and_die("unrecognized file type");
		}
	}

	if (!(archive_handle->ah_flags & ARCHIVE_NOPRESERVE_OWN)) {
#if ENABLE_FEATURE_TAR_UNAME_GNAME
		uid_t uid = file_header->uid;
		gid_t gid = file_header->gid;

		if (file_header->uname) {
			struct passwd *pwd = getpwnam(file_header->uname);
			if (pwd) uid = pwd->pw_uid;
		}
		if (file_header->gname) {
			struct group *grp = getgrnam(file_header->gname);
			if (grp) gid = grp->gr_gid;
		}
		lchown(file_header->name, uid, gid);
#else
		lchown(file_header->name, file_header->uid, file_header->gid);
#endif
	}
	if ((file_header->mode & S_IFMT) != S_IFLNK) {
		/* uclibc has no lchmod, glibc is even stranger -
		 * it has lchmod which seems to do nothing!
		 * so we use chmod... */
		if (!(archive_handle->ah_flags & ARCHIVE_NOPRESERVE_PERM)) {
			chmod(file_header->name, file_header->mode);
		}
		/* same for utime */
		if (archive_handle->ah_flags & ARCHIVE_PRESERVE_DATE) {
			struct utimbuf t;
			t.actime = t.modtime = file_header->mtime;
			utime(file_header->name, &t);
		}
	}
}
